#
# AOLServer tcl emulation --- Copyright Robert S. Thau, 2000.

# This is the bootstrap script which sets up the tcl environment for
# aolsever emulation.  It is run by a master interpreter which actually
# sets up a slave interpreter which runs the aolserver code.  That's
# done so that we can implement stuff here which is implemented in
# the C code of the actual aolserver, without allowing aolserver tcl
# code per se to mess with it.

# However, to do that, we use a few moderately dangerous commands which
# are not available to aolserver tcl generally; also, the C code relies
# on code here to do some housekeeping which can crash the server if
# done wrong.  So, don't be too quick to mess with this stuff unless
# you are comfortable running the server under a C debugger if need be.

set initializing 1;  # Flag saying whether we're running init code...

# A lot of the aolserver emulation here is actually written in tcl.
# More should be --- the amount of library reinvention here is
# staggering.  Load in some basic support routines, and then the
# code that implements the various APIs...

proc load_bootstrap_tcl {file} {
    global ap_server_conf
    set fn [file join $ap_server_conf(SERVER_ROOT) libexec aol_bootstrap $file]
    ns_log notice "Loading bootstrap support: $fn"
    eval [list uplevel [list source $fn]]
}

# Load basic support routines

load_bootstrap_tcl Init_hooks.tcl;	# Define hooks and basic machinery
load_bootstrap_tcl Init_slave.tcl;	# Initialize the slave interpreter

# Load APIs and other internal support

load_bootstrap_tcl apache.tcl;		# Dealing with low-level Apache stuff
load_bootstrap_tcl shared_vars.tcl;	# ns_share and nsv 
load_bootstrap_tcl files.tcl;		# channels and file-related commands
load_bootstrap_tcl status.tcl;		# ns_status, ns_info, ns_conn, etc.
load_bootstrap_tcl config.tcl;		# ns_config cmds, config file handling
load_bootstrap_tcl regproc.tcl;		# ns_register_proc, etc.
load_bootstrap_tcl regfilter.tcl;	# ns_register_filter, etc.
load_bootstrap_tcl sched.tcl;		# ns_schedule_foo, etc.
load_bootstrap_tcl dbsupport.tcl;	# database support
load_bootstrap_tcl misc.tcl;		# miscellany

################################################################
#
# So much for loading support.  Now to actually read our config
# files and get initialized.

# The server name is tied up with the main config file in an
# ugly way.  In real AOLserver, you can specify the server name
# on the command line with the -s flag.  If you don't, then the
# server name is undefined (and you'll probably get a SIGSEGV
# if you call [ns_info server]) until after the main config file
# is read.  At that point, if the server name wasn't set on the
# command line, then AOLserver verifies that the ns/servers section
# has a single entry and uses it as the server name.

# We emulate this process by allowing the server name to be set
# in the Apache config file instead of via a -s command line flag.

set server_name ""

if [info exists ap_server_conf(server_name)] {
    # Server name was set in Apache config.
    set server_name $ap_server_conf(server_name)
    ns_set_servername $server_name;	# synch up the C code.
}

# Load main aolserver config file

set config_filename [ap_config_filename config_file conf/nsd.ini]

if [string match *.ini $config_filename] {
    read_ini_file $config_filename
} else {
    source $config_filename
}

if {[string compare $server_name ""] == 0} {
    # Server name not yet set...

    set serversec [ns_configsection ns/servers]
    switch -- [ns_set size $serversec] {
      1 {
	  set server_name [ns_set key $serversec 0]
	  ns_set_servername $server_name;	# synch up the C code.
      }
      0 {
	error "Missing ns/servers config section: cannot determine server name"
      }
      default {
	error "More than one entry in ns/servers config section: you must specify a server name in the Apache configuration."
      }
    }
}

# Load all tcl_modules (and also .ini files in that directory)

load_modules_from_dir [ap_config_filename tcl_modules tcl_modules]

# Now handle the auxconfigdir, if specified...

set auxdir [ns_config ns/parameters AuxConfigDir]
ns_log Notice "Aux config dir is \"$auxdir\""
if [string compare $auxdir ""] { load_modules_from_dir $auxdir }

# Now that we've loaded all the parameter files, we can set
# up the database system...

db_init_pools

# Load the user's Tcl library.

set auxdir [ns_info tcllib]
ns_log Notice "Tcl library dir is \"$auxdir\""
if [string compare $auxdir ""] { load_modules_from_dir $auxdir }
unset auxdir

# Run initialization scripts (init code which ACS does as
# "ns_schedule_proc -once"s because it can't talk to the database
# while configuring itself in the real aolserver).  We must do
# this after initializing the db; this ordering requirment is
# why I haven't just made db_init_pools and run_init_scripts
# entries on some hook list (which would hide the dependancy).

run_init_scripts

# After running the init scripts, we are now in the final configuration
# for running code in child processes.  So, we reset the initializing flag to
# disallow changes which would allow the child processes to get out of synch,
# and run our final setup...

set initializing 0

run_hooks final_setup



