################################################################
#
# Dealing with files

# Part of the cleanup is closing any filehandles left open by
# tcl code (which can easily happen --- open a file, do something
# that throws an error, and it never executes your close statement).
# So, to avoid potentially nasty leaks, we override the definitions
# of commands that open and close channels in the slave interpreter
# with variants that let us do a little accounting:

lappend per_req_vars req_channels

proc slave_channel_cleanup {} {
    global req_channels slaveInterp
    if [info exists req_channels] {
	foreach chan $req_channels {
	    catch {$slaveInterp invokehidden close $chan}
	}
    }
}

run_at cleanup slave_channel_cleanup

$slaveInterp hide open
$slaveInterp hide socket
$slaveInterp hide ns_sockopen
$slaveInterp hide close

proc safe_apply_hidden { cmd cmd_args } {
    global slaveInterp

    # The following would be slightly less awkward if tcl's lreplace
    # allowed inserting elements at the beginning or end of a list,
    # though the eval is kinda hideous regardless.
    
    set cmd [list $slaveInterp invokehidden $cmd]
    foreach arg $cmd_args { lappend cmd $arg }
    return [eval $cmd]
}

proc acc_open {args} {
    global req_channels slaveInterp
    set handle [safe_apply_hidden open $args]

    # AOLserver has a hacked Tcl that does this automatically.
    $slaveInterp eval fconfigure $handle -translation binary

    lappend req_channels $handle
    return $handle
}

proc acc_socket {args} {
    global req_channels slaveInterp
    set handle [safe_apply_hidden socket $args]

    # AOLserver has a hacked Tcl that does this automatically.
    $slaveInterp eval fconfigure $handle -translation binary

    lappend req_channels $handle
    return $handle
}

proc acc_ns_sockopen {args} {
    global req_channels slaveInterp
    set handles [safe_apply_hidden ns_sockopen $args]
    foreach handle $handles {
	# AOLserver has a hacked Tcl that does this automatically.
	$slaveInterp eval fconfigure $handle -translation binary

	lappend req_channels $handle
    }
    return $handles
}

proc acc_close {handle} {
    global req_channels
    set posn [lsearch -exact $req_channels $handle]
    if {$posn >= 0} {set req_channels [lreplace $req_channels $posn $posn]}
    return [safe_apply_hidden close $handle]
}

$slaveInterp alias open        acc_open
$slaveInterp alias socket      acc_socket
$slaveInterp alias ns_sockopen acc_ns_sockopen
$slaveInterp alias close       acc_close

# We also define here some file-related AolServer tcl functions.
# If these were used to send large files, this would get really
# wasteful.  Fortunately, they're not...

$slaveInterp eval {

    # Aren't these two kinda redundant?
    
    proc ns_writefp {chan args} {
    
	if {[llength $args] > 0} {
	    set len [lindex $args 0]
	    set stuff [read $chan $len]
	} else {
	    set stuff [read $chan]
	}

	ns_write $stuff
    }

    proc ns_connsendfp {args} {
	if {[llength $args] == 3} { set $args [lrange $args 1 end] }
	if {[llength $args] != 2} {
	    error "Bad args --- ns_connsendfp ?conn? chan length"
	}
	ns_writefp [lindex $args 0] [lindex $args 1]
    }
}

