################################################################
#
# ns_register_proc

# The tree data structure we use for these...
# Note that we *don't* bother mirroring the directory structure here;
# you'd have to be registering several dozen procs at least before that game
# was close to worth the candle.  Keys are just (type method dir) triples
# (though we forgo type for exact matches of course)...
#
# and we sort type and dir by length to be sure we go from most to least
# specific...

set wild_procs ""
set exact_procs ""

proc replace_in_tree { tree val keys } {

    # Base case --- if no selectors, we're replacing the whole tree
    
    if {[llength $keys] == 0} { return $val }

    # Find and replace matching branch, if there is one...

    set keyfirst [lindex $keys 0]
    set keyrest  [lrange $keys 1 end]
    set idx 0
    
    foreach subtree $tree {
	if ![string compare [lindex $subtree 0] $keyfirst] {
	    set branch [ lindex [ lindex $tree $idx ] 1]
	    set newbranch [replace_in_tree $branch $val $keyrest]
	    return [lreplace $tree $idx $idx [ list $keyfirst $newbranch]]
	}
	incr idx
    }

    # There isn't.  Add it.

    lappend tree [list $keyfirst [replace_in_tree [list] $val $keyrest]]
    return $tree
}

# To be sure we match most-specific first, we sort the tree so that
# at each level, branches with the longest keys go first.

proc comp_key_lengths { l1 l2 } {
    return [expr [string length [lindex $l1 0]]-[string length [lindex $l2 0]]]
}

proc sort_tree_branches { tree levels } {
    if {$levels == 0} { return $tree }

    # First, sort lower levels...
    
    if {$levels > 1} {
	set nlevels [expr $levels - 1]
	set newtree ""
	foreach branch $tree {
	    set key [lindex $branch 0]
	    set branch [lindex $branch 1]
	    lappend newtree [ list $key [sort_tree_branches $branch $nlevels ]]
	}
	set tree $newtree
    }

    # Now, sort at this level

    return [lsort -command comp_key_lengths $tree]
}

proc finalize_registered_procs {} {
    global exact_procs wild_procs
    set exact_procs [ sort_tree_branches $exact_procs 2 ]
    set wild_procs [ sort_tree_branches $wild_procs 3 ]
}

run_at final_setup finalize_registered_procs

# Matching in the tree 

proc match_tree_inexact { tree keys } {
    
    if {[llength $keys] == 0} { return $tree }

    set keyfirst [lindex $keys 0]
    set keyrest  [lrange $keys 1 end]

    foreach subtree $tree {
	if [string match [lindex $subtree 0] $keyfirst] {
	    set maybe [match_tree_inexact [lindex $subtree 1] $keyrest]
	    if {[string length $maybe] > 0} {return $maybe}
	}
    }

    return ""
}

# Finding matching branches in the trees

proc match_tree_exact { tree keys } {

    if {[llength $keys] == 0} { return $tree }

    set keyfirst [lindex $keys 0]
    set keyrest  [lrange $keys 1 end]
    
    foreach subtree $tree {
	if ![string compare [lindex $subtree 0] $keyfirst] {
	    set maybe [match_tree_exact [lindex $subtree 1] $keyrest]
	    if {[string length $maybe] > 0} {return $maybe}
	}
    }

    return ""
}

proc match_tree { tree args } {
    set keyfirst [lindex keys 0]
    set keyrest  [lrange keys 1 end]
    
    foreach subtree $tree {
	if ![string match [lindex $subtree 0] $keyfirst] {
	    return [match_tree [lindex $subtree 1] $keyrest]
	}
    }

    return ""
}

# Running registered procs...

proc run_registered_proc {regproc} {

    global slaveInterp
    
    set pproc [lindex $regproc 0]
    set arg [lindex $regproc 1]
    set nargs 0

    # Don't gripe now if the function's not defined; we'll gripe later
    # when we try to call it.
    
    catch {set nargs [llength [$slaveInterp eval [list info args $pproc]]]} 
    
    # You know, some people might think that wiring indentation style
    # into the syntax of a language is a bad idea...

    if {$nargs == 0} {       set cmd [list $pproc]
    } elseif {$nargs == 1} { set cmd [list $pproc $arg]
    } else                 { set cmd [list $pproc conn $arg] }

    return [safe_eval $cmd registered_proc ""]
}

proc run_registered_procs {method url} {

    global exact_procs wild_procs
    
    # Search first for an exact match

    set exmatch [match_tree_exact $exact_procs [list $method $url]]

    if {[string length $exmatch] > 0} {
	run_registered_proc $exmatch
	return 1
    }

    # If not, parse out last component, and go wild

    set lastslash [string last / $url]
    set lastcomponent [string range $url [expr $lastslash + 1] end]

    set match [match_tree_inexact $wild_procs \
		   [list $lastcomponent $method $url/]]
    
    if {[string length $match] > 0} {
	run_registered_proc $match
	return 1
    }

    return 0
}

# Finally, registering a proc

proc register_proc {args} {

    global exact_procs wild_procs initializing
    set exact 0

    # Are we are allowing new procs?

    if !$initializing {
	error "Can't register procs once server is configured"
    }
    
    # Parse arguments
    
    if ![string compare [lindex $args 0] -noinherit] {
	set exact 1
	set args [lrange $args 1 end]
    }

    set method [lindex $args 0]
    set url    [lindex $args 1]
    set pproc  [lindex $args 2]
    set arg    ""

    if {[llength $args] > 3} { set arg [lindex $args 3] }

    # Check out the URL, and separate it into "file" and "directory" components
    
    if [string compare [string range $url 0 0] /] {
	error "Url $url does not start with / in ns_register_proc"
    }

    set lastslash [string last / $url]
    set filemask [string range $url [expr $lastslash + 1] end]

    if $exact {
	set dirstring $url
    } elseif {[string first * $filemask]<0 && [string first ? $filemask]<0} {
	# No wildcard
	set filemask *
	set dirstring "$url/*"
    } else {
	set dirstring [string range $url 0 [expr $lastslash - 1]]/*
    }

    regsub -all {//+} $dirstring / dirstring

    # And actually register the thing... 

    set procstuff [list $pproc $arg]

    if $exact {
	set exact_procs \
	    [ replace_in_tree $exact_procs $procstuff \
		  [list $method $dirstring ] ]
    } else {
	set wild_procs \
	    [ replace_in_tree $wild_procs $procstuff \
		  [list $filemask $method $dirstring ] ]
    }
}

$slaveInterp alias ns_register_proc register_proc

